import os, sys

def exec_shell(cmd, print_result = False):
    with os.popen(cmd) as pf:
        lines = pf.readlines()
    if print_result:
        for line in lines:
            print(line.strip())
    return ''.join(lines)

GCC_CXX_COMPILER = "g++"
CLANG_CXX_COMPILER = "clang++"

class dynamic_obj:
    def __init__(self, file, is_str = False):
        if not is_str:
            with open(file, "rt") as f:
                lines = f.readlines()
                self.code = ''.join(lines)
        else:
            self.code = file
        script_dir = os.path.dirname(os.path.abspath(__file__))
        self.compile_opts = [f'-I{script_dir}/include', '-lcuda_mock_impl', f'-L{script_dir}', f'-Wl,-rpath,{script_dir}']
    def compile(self):
        if not self.check_env():
            raise Exception("can't find host compiler!")
        from datetime import datetime
        current_time = datetime.now()
        current_time = str(current_time).replace(':', '_')
        current_time = str(current_time).replace(' ', '_')
        current_time = str(current_time).replace('.', '_')
        
        file_name = f'__cuda_rt_hook__{current_time}'
        cpp_file_name = f'/tmp/{file_name}.cpp'

        with open(cpp_file_name, 'wt') as f:
            f.write(self.code)

        self.lib_file_name = f'/tmp/{file_name}.so'

        opts = ' '.join(self.compile_opts)
        compile_cmd = f'{self.compiler} -shared -fpic {cpp_file_name} {opts} -o {self.lib_file_name}'
        exec_shell(compile_cmd, True)
        return self

    def check_env(self):
        if len(self.get_compiler(GCC_CXX_COMPILER)) == 0:
            self.compiler = GCC_CXX_COMPILER
            return True
        if len(self.get_compiler(CLANG_CXX_COMPILER)) == 0:
            self.compiler = CLANG_CXX_COMPILER
            return True
        return False
        
    def get_compiler(self, name):
        result = exec_shell(f'which {name}').strip()
        if 0 == len(result):
            return ''
        if -1 == result.find("not found"):
            return ''
        return result

    def appen_compile_opts(self, *args):
        self.compile_opts.extend(list(args))
        return self
    
    def get_lib(self):
        return self.lib_file_name


import os
import ctypes
from .dynamic_obj import *

script_dir = os.path.dirname(os.path.abspath(__file__))
cuda_mock_impl = ctypes.CDLL(f'{script_dir}/libcuda_mock_impl.so')

def add(lhs, rhs):
    return lhs + rhs

def initialize():
    return cuda_mock_impl.initialize()

def uninitialize():
    return cuda_mock_impl.uninitialize()

def internal_install_hook(*args):
    new_args = [ctypes.c_char_p(arg.encode('utf-8')) for arg in args]
    return cuda_mock_impl.internal_install_hook(*new_args)

def internal_install_hook_regex(*args):
    new_args = [ctypes.c_char_p(arg.encode('utf-8')) for arg in args]
    return cuda_mock_impl.internal_install_hook_regex(*new_args)

def xpu_initialize():
    return cuda_mock_impl.xpu_initialize()

str_code = '''
#include <stddef.h>
#include <stdio.h>
#include <unistd.h>

#include <fstream>
#include <iostream>

#include "logger/logger.h"
#include "profile/Timer.h"

#define EXPORT __attribute__((__visibility__("default")))

extern "C" {{

void* __origin_{0} = nullptr;
void* __origin_{1} = nullptr;
void* __origin_{2} = nullptr;


EXPORT int {0}(const void* arg, uint64_t size, uint64_t offset) {{
    hook::Timer timer(__func__);
    int ret = reinterpret_cast<decltype(&{0})>(__origin_{0})(arg, size, offset);
    return ret;
}}

EXPORT int {1}(void* func) {{
    hook::Timer timer(__func__);
    int ret = reinterpret_cast<decltype(&{1})>(__origin_{1})(func);
    return ret;
}}

EXPORT int {2}(int nclusters, int ncores, void* stream) {{
    hook::Timer timer(__func__);
    int ret = reinterpret_cast<decltype(&{2})>(__origin_{2})(nclusters, ncores, stream);
    return ret;
}}

}}
'''

def xpu_runtime_profile():
    script_dir = os.path.dirname(os.path.abspath(__file__))
    funcs = ["xpu_launch_argument_set", "xpu_launch_async", "xpu_launch_config"]
    lib = dynamic_obj(str_code.format(*funcs), True).appen_compile_opts('-lpthread').compile().get_lib()
    internal_install_hook_regex(r"libxpurt\.so\..*", r"[^(libxpurt\.so\..*)]", "xpu_launch_argument_set", str(lib), "xpu_launch_argument_set")
    internal_install_hook_regex(r"libxpurt\.so\..*", r"[^(libxpurt\.so\..*)]", "xpu_launch_async", str(lib), "xpu_launch_async")
    internal_install_hook_regex(r"libxpurt\.so\..*", r"[^(libxpurt\.so\..*)]", "xpu_launch_config", str(lib), "xpu_launch_config")





from setuptools import setup, Extension, find_packages
from setuptools.command.build_ext import build_ext
import sys
import setuptools
import os
import subprocess
from setuptools.extension import Extension as _Extension
from glob import glob
import shutil

class CMakeExtension(_Extension):
    def __init__(self, name, sourcedir=''):
        super().__init__(name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)

class CMakeBuild(build_ext):
    def run(self):
        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        cmake_args = [
            '-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name))),
            '-DPYTHON_EXECUTABLE=' + sys.executable,
            '-DCMAKE_INSTALL_PREFIX=' + os.path.join(os.path.abspath(os.path.dirname(self.build_lib)), os.path.basename(self.build_lib))
        ]

        build_dir = os.path.abspath(os.path.join(self.build_temp, ext.name))
        os.makedirs(build_dir, exist_ok=True)

        ninja_args = [
            '-GNinja',  # Specify Ninja as the generator
        ]

        subprocess.check_call(['cmake', ext.sourcedir] + cmake_args + ninja_args, cwd=build_dir)
        subprocess.check_call(['cmake', '--build', '.'], cwd=build_dir)
        subprocess.check_call(['ninja', 'install'], cwd=build_dir)

setup(
    name="cuda-mock",
    version="0.0.8",
    author="lipracer",
    author_email="lipracer@gmail.com",
    description="a tools hook some api call at runtime",

    url="https://github.com/lipracer/torch-cuda-mock", 

    ext_modules=[CMakeExtension('cuda_mock_impl', sourcedir='.')],
    cmdclass=dict(build_ext=CMakeBuild),
    package_dir={'': 'src'},
    package_data={'' : ['*']},
    zip_safe=False,
)
